from __future__ import unicode_literals

import locale
import os, os.path
import time
from datetime import datetime
from language import get_string
import urllib.request, urllib.parse, urllib.error
from kodi_six import xbmc, xbmcgui

DATE_FORMAT = "%Y-%m-%d %H:%M"
MONTH_NAMES = []

if not MONTH_NAMES:
    MONTH_NAMES.append(get_string("January"))
    MONTH_NAMES.append(get_string("February"))
    MONTH_NAMES.append(get_string("March"))
    MONTH_NAMES.append(get_string("April"))
    MONTH_NAMES.append(get_string("May"))
    MONTH_NAMES.append(get_string("June"))
    MONTH_NAMES.append(get_string("July"))
    MONTH_NAMES.append(get_string("August"))
    MONTH_NAMES.append(get_string("September"))
    MONTH_NAMES.append(get_string("October"))
    MONTH_NAMES.append(get_string("November"))
    MONTH_NAMES.append(get_string("Dezember"))


def log(x):
    xbmc.log(repr(x), xbmc.LOGERROR)


def windows():
    if os.name == 'nt':
        return True
    else:
        return False


def escape(value):
    try:
        value = value.decode()
    except:
        pass
    try:
        value = value.encode()
    except:
        pass
    return urllib.parse.quote_plus(value)


def get_format():
#    dateFormat = xbmc.getRegion('datelong')
    dateFormat = xbmc.getRegion('dateshort')
    timeFormat = xbmc.getRegion('time').replace('%H%H', '%H').replace('%I%I', '%I')
    timeFormat = timeFormat.replace(":%S", "")
    return "{} {}".format(dateFormat, timeFormat)


def extract_date(dateLabel, timeLabel):
    dateString = xbmc.getInfoLabel(dateLabel)
    timeString = xbmc.getInfoLabel(timeLabel)

    dateParts = dateString.split(" ")

    dayString = dateParts[1].replace(".","")
    if int(dayString) < 10:
        dayString = "0" + dayString

    monthString = dateParts[2].replace(" ", "")
    monthNumber = MONTH_NAMES.index(monthString) + 1
    if monthNumber < 10:
        monthString = "0" + str(monthNumber)
    else:
        monthString = str(monthNumber)

    yearString = dateParts[3].replace(" ", "")

    fullDate = "{}.{}.{} {}".format(dayString, monthString, yearString, timeString)

    #log("fullDate: " + fullDate)

    try:
        parsedDate = datetime.strptime(fullDate, fullFormat)
    except TypeError:
        parsedDate = datetime(*(time.strptime(fullDate, fullFormat)[0:6]))

    """
		#orininal code
    fullDate = "{}, {}".format(dateString, timeString)
    #log("fullDate: " + fullDate)

    # https://bugs.python.org/issue27400
    try:
        parsedDate = datetime.strptime(fullDate, fullFormat)
    except TypeError:
        parsedDate = datetime(*(time.strptime(fullDate, DATE_FORMAT)[0:6]))
    """

    return datetime.strftime(parsedDate, DATE_FORMAT)


def get_language():
    try:
        language = xbmc.getLanguage(xbmc.ISO_639_1, True)
        languageParts = language.split("-")
        return "{}_{}.UTF-8".format(languageParts[0], languageParts[1])
    except:
        return ""


#try:
#    usedLocale = locale.setlocale(locale.LC_TIME, get_language())
#except:
#    usedLocale = locale.setlocale(locale.LC_TIME, "")
#log("Used locale: " + usedLocale)

fullFormat = get_format()

channel = escape(xbmc.getInfoLabel("ListItem.ChannelName"))
title = escape(xbmc.getInfoLabel("ListItem.Label"))

try:
    start = extract_date("ListItem.StartDate", "ListItem.StartTime")
    stop = extract_date("ListItem.EndDate", "ListItem.EndTime")

    try:
        cmd = "PlayMedia(plugin://plugin.video.iptv.recorder/record_epg/%s/%s/%s/%s)" % (channel, title, start, stop)
        xbmc.executebuiltin(cmd)
        message = "{}: {} ({} {} {})'".format(xbmc.getInfoLabel("ListItem.ChannelName"), xbmc.getInfoLabel("ListItem.Label"), start, get_string("to"), stop)
        xbmcgui.Dialog().notification("IPTV Recorder - " + get_string("Scheduled record"), message, xbmcgui.NOTIFICATION_INFO, 10000, sound=False)
    except:
        xbmcgui.Dialog().notification("IPTV Recorder", "Could not schedule recording", xbmcgui.NOTIFICATION_WARNING)
except Exception as e:
    xbmcgui.Dialog().notification("IPTV Recorder", "Error parsing dates", xbmcgui.NOTIFICATION_ERROR)
    log("IPTV Recorder: Error parsing dates ({})".format(e))

if MONTH_NAMES:
    del MONTH_NAMES
